"""Run gbar on the first available display from a list of displays."""

import asyncio
import contextlib
from time import time

from ..common import state
from .interface import Plugin


class Extension(Plugin):
    """Manage gBar application."""

    monitors: set[str]
    proc = None
    cur_monitor: str | None = ""

    ongoing_task: asyncio.Task | None = None

    async def on_reload(self) -> None:
        """Start the process."""
        self.kill()
        self._run_gbar()

    def _run_gbar(self) -> None:
        """Create ongoing task restarting gbar in case of crash."""
        if self.ongoing_task:
            self.ongoing_task.cancel()

        async def _run_loop() -> None:
            prev_time = time()
            while True:
                await self.set_best_monitor()
                cmd = f"gBar bar {self.cur_monitor}"
                now = time()
                self.proc = await asyncio.create_subprocess_shell(cmd)
                await self.proc.wait()
                delay = 60 - (now - prev_time)
                text = f"gBar crashed, restarting in {delay // 2}s." if delay > 0 else "gBar crashed, restarting."
                await self.notify_error(text)
                prev_time = now
                if delay > 0:
                    await asyncio.sleep(delay / 2)

        self.ongoing_task = asyncio.create_task(_run_loop())

    async def run_gbar(self, args: str) -> None:
        """Start gBar on the first available monitor."""
        if args.startswith("re"):
            self.kill()
            await self.on_reload()

    async def set_best_monitor(self) -> None:
        """Set the best monitor to use in `cur_monitor`."""
        self.cur_monitor = await self.get_best_monitor()
        if not self.cur_monitor:
            self.cur_monitor = next(iter(state.monitors))
            await self.notify_info(f"gBar: No preferred monitor found, using {self.cur_monitor}")

    async def get_best_monitor(self) -> str:
        """Get best monitor according to preferred list."""
        preferred: list[str] = self.config.get("monitors", [])
        monitors = [m for m in await self.hyprctl_json("monitors") if m.get("currentFormat") != "Invalid"]
        names = [m["name"] for m in monitors]
        for monitor in preferred:
            if monitor in names:
                return monitor
        return ""

    async def event_monitoradded(self, monitor: str) -> None:
        """Switch bar in case the monitor is preferred."""
        if self.cur_monitor:
            preferred = self.config.get("monitors", [])
            cur_idx = preferred.index(self.cur_monitor) if self.cur_monitor else 999
            new_idx = preferred.index(monitor)
            if 0 <= new_idx < cur_idx:
                self.kill()
                await self.on_reload()

    async def exit(self) -> None:
        """Kill the process."""
        self.kill()

    def kill(self) -> None:
        """Kill the process."""
        if self.proc:
            if self.ongoing_task:
                self.ongoing_task.cancel()
                self.ongoing_task = None
            with contextlib.suppress(ProcessLookupError):
                self.proc.kill()
            self.proc = None



"""Conversion functions for units used in Pyprland & plugins."""

from ..common import is_rotated
from ..types import MonitorInfo


def convert_monitor_dimension(size: int | str, ref_value: int, monitor: MonitorInfo) -> int:
    """Convert `size` into pixels (given a reference value applied to a `monitor`).

    if size is an integer, assumed pixels & return it
    if size is a string, expects a "%" or "px" suffix
    else throws an error
    """
    if isinstance(size, int):
        return size

    if isinstance(size, str):
        if size.endswith("%"):
            p = int(size[:-1])
            return int(ref_value / monitor["scale"] * p / 100)
        if size.endswith("px"):
            return int(size[:-2])

    msg = f"Unsupported format: {size} (applied to {ref_value})"
    raise ValueError(msg)


def convert_coords(coords: str, monitor: MonitorInfo) -> list[int]:
    """Convert a string like "X Y" to coordinates relative to monitor.

    Supported formats for X, Y:
    - Percentage: "V%". V in [0; 100]
    - Pixels: "Vpx". V should fit in your screen and not be zero

    Example:
    "10% 20%", monitor 800x600 => 80, 120
    """
    return [
        convert_monitor_dimension(name, monitor[ref], monitor)  # type: ignore
        for (name, ref) in zip(
            [coord.strip() for coord in coords.split()],
            (("height", "width") if is_rotated(monitor) else ("width", "height")),
            strict=False,
        )
    ]



"""Add system notifications based on journal logs."""

import asyncio
import contextlib
import re
from copy import deepcopy
from typing import cast

from ..adapters.colors import convert_color
from ..common import apply_filter
from .interface import Plugin

try:
    from asyncio.subprocess import Process
except ImportError:
    from asyncio import Process  # type: ignore

builtin_parsers = {
    "journal": [
        {
            "pattern": r"([a-z0-9]+): Link UP$",
            "filter": r"s/.*\[\d+\]: ([a-z0-9]+): Link.*/\1 is active/",
            "color": "#00aa00",
        },
        {
            "pattern": r"([a-z0-9]+): Link DOWN$",
            "filter": r"s/.*\[\d+\]: ([a-z0-9]+): Link.*/\1 is inactive/",
            "color": "#ff8800",
        },
        {
            "pattern": r"Process \d+ \(.*\) of .* dumped core.$",
            "filter": r"s/.*Process \d+ \((.*)\) of .* dumped core./\1 dumped core/",
            "color": "#aa0000",
        },
        {
            "pattern": r"usb \d+-[0-9.]+: Product: ",
            "filter": r"s/.*usb \d+-[0-9.]+: Product: (.*)/USB plugged: \1/",
        },
    ]
}


class Extension(Plugin):
    """Notification system from live apps & logs."""

    def __init__(self, name: str) -> None:
        """Initialize the class."""
        super().__init__(name)
        self.tasks: list[asyncio.Task] = []
        self.sources: dict[str, Process] = {}
        self.parsers: dict[str, asyncio.Queue] = {}
        self.running = True

    async def on_reload(self) -> None:
        """Reload the plugin."""
        await self.exit()
        self.running = True
        parsers = deepcopy(builtin_parsers)
        parsers.update(self.config.get("parsers", {}))
        for name, pprops in parsers.items():
            self.tasks.append(asyncio.create_task(self.start_parser(name, pprops)))
            self.parsers[name] = asyncio.Queue()
            self.log.debug("Loaded parser %s", name)

        for props in self.config.get("sources", []):
            assert props["parser"] in self.parsers, f"{props['parser']} was not found in {self.parsers}"
            self.log.debug("Loaded source %s => %s", props["command"], props["parser"])
            self.tasks.append(asyncio.create_task(self.start_source(props)))

    async def exit(self) -> None:
        """Exit function."""
        self.running = False
        for task in self.tasks:
            task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await task
        for source in self.sources.values():
            if source.pid is not None:
                source.kill()
        self.tasks[:] = []

    async def start_source(self, props: dict[str, str]) -> None:
        """Start a source loop.

        A source is a command that will be executed and its stdout will be read line by line.

        Args:
            props: A dictionary with the following keys:
                - command: The command to execute.
                - parser: The name of the parser to use.
        """
        parsers = [props["parser"]] if isinstance(props["parser"], str) else props["parser"]
        queues = [self.parsers[p] for p in parsers]
        proc = await asyncio.create_subprocess_shell(props["command"], stdout=asyncio.subprocess.PIPE)

        self.sources[props["command"]] = proc
        assert proc.stdout
        await asyncio.sleep(1)
        # Read stdout line by line and push to parser
        while proc.returncode is None and self.running:
            line = (await proc.stdout.readline()).decode().strip()
            if line:
                for q in queues:
                    await q.put(line)

    async def start_parser(self, name: str, props: list) -> None:
        """Start a parser loop.

        Args:
            name: The name of the parser.
            props: A list of dictionaries with the following keys:
                - pattern: A regex pattern to match.
                - filter: A filter to apply to the matched line.
                - color: The color to use in the notification.
        """
        q = self.parsers[name]
        default_color = self.config.get("default_color", "#5555AA")
        rules = [
            {
                "pattern": re.compile(prop["pattern"]),
                "filter": prop.get("filter"),
                "color": convert_color(prop.get("color", default_color)),
            }
            for prop in props
        ]
        while self.running:
            content = await q.get()
            for rule in rules:
                if rule["pattern"].search(content):
                    text = apply_filter(content, cast(str, rule["filter"])) if rule["filter"] else content
                    await self.notify(text, color=rule["color"])
                    await asyncio.sleep(0.01)



"""The monitors plugin."""

import asyncio
from collections import defaultdict
from copy import deepcopy
from typing import Any, cast

from ..common import CastBoolMixin, is_rotated, state
from ..types import MonitorInfo
from .interface import Plugin

MONITOR_PROPS = {"scale", "transform", "rate", "resolution"}


def trim_offset(monitors: list[MonitorInfo]) -> None:
    """Make the monitor set layout start at 0,0."""
    off_x = None
    off_y = None
    for mon in monitors:
        if off_x is None:
            off_x = mon["x"]

        if off_y is None:
            off_y = mon["y"]

        off_x = min(mon["x"], off_x)
        off_y = min(mon["y"], off_y)

    assert off_x is not None
    assert off_y is not None

    for mon in monitors:
        mon["x"] -= off_x
        mon["y"] -= off_y


def clean_pos(position: str) -> str:
    """Harmonize position format."""
    return position.lower().replace("_", "").replace("-", "")


def scale_and_rotate_mon(monitor: MonitorInfo) -> tuple[int, int]:
    """Scale and rotate the monitor dimensions."""
    width = int(monitor["width"] / monitor["scale"])
    height = int(monitor["height"] / monitor["scale"])
    if is_rotated(monitor):
        width, height = height, width
    return width, height


def get_xy(place: str, main_mon: MonitorInfo, other_mon: MonitorInfo) -> tuple[int, int]:
    """Get the XY position of a monitor according to another (after `place` is applied).

    Place syntax: "<top|left|bottom|right> [center|middle|end] of" (without spaces)
    """
    align_x = False  # if alignment is on X axis, else on Y axis
    scaled_m_w, scaled_m_h = scale_and_rotate_mon(main_mon)
    scaled_om_w, scaled_om_h = scale_and_rotate_mon(other_mon)

    if place[0] in ("t", "b"):  # top or bottom
        align_x = True
        x = other_mon["x"]
        y = other_mon["y"] - scaled_m_h if place[0] == "t" else other_mon["y"] + scaled_om_h
    else:  # left or right
        y = other_mon["y"]
        x = other_mon["x"] - scaled_m_w if place[0] == "l" else other_mon["x"] + scaled_om_w

    centered = "middle" in place or "center" in place

    if align_x:
        if centered:
            x += int((scaled_om_w - scaled_m_w) / 2)
        elif "end" in place:
            x += int(scaled_om_w - scaled_m_w)
    elif centered:
        y += int((scaled_om_h - scaled_m_h) / 2)
    elif "end" in place:
        y -= scaled_m_h - scaled_om_h

    return (x, y)


def build_graph(config: dict[str, dict[str, list[str]]]) -> dict[str, list[str]]:
    """Make a sorted graph based on the cleaned_config."""
    graph = defaultdict(list)
    for name1, positions in config.items():
        for pos, names in positions.items():
            if pos in MONITOR_PROPS:
                continue
            tldr_direction = pos.startswith(("left", "top"))
            for name2 in names:
                if tldr_direction:
                    graph[name1].append(name2)
                else:
                    graph[name2].append(name1)
    return graph


class Extension(CastBoolMixin, Plugin):  # pylint: disable=missing-class-docstring
    """Control monitors layout."""

    _mon_by_pat_cache: dict[str, MonitorInfo] = {}

    async def on_reload(self) -> None:
        """Reload the plugin."""
        self._clear_mon_by_pat_cache()
        monitors = await self.hyprctl_json("monitors")

        for mon in state.monitors:
            await self._hotplug_command(name=mon, monitors=monitors)

        if self.cast_bool(self.config.get("startup_relayout"), True):
            # run relayout after 1second without blocking

            async def _delayed_relayout() -> None:
                await self._run_relayout(monitors)
                await asyncio.sleep(1)
                await self._run_relayout()

            await asyncio.create_task(_delayed_relayout())

    def _build_monitor_command(self, monitor: MonitorInfo, config: dict[str, dict[str, Any]], every_monitor: dict[str, MonitorInfo]) -> str:
        """Build the monitor command."""
        name = monitor["name"]
        this_config = config.get(name, {})
        rate = this_config.get("rate", every_monitor[name]["refreshRate"])
        res = this_config.get("resolution", f"{every_monitor[name]['width']}x{every_monitor[name]['height']}")
        if isinstance(res, list):
            res = f"{res[0]}x{res[1]}"
        scale = this_config.get("scale", every_monitor[name]["scale"])
        position = f"{monitor['x']}x{monitor['y']}"
        transform = this_config.get("transform", every_monitor[name]["transform"])
        return f"monitor {name},{res}@{rate},{position},{scale},transform,{transform}"

    # Command

    async def run_relayout(self) -> bool:
        """Recompute & apply every monitors's layout."""
        return await self._run_relayout()

    async def _run_relayout(self, monitors: list[MonitorInfo] | str | None = None) -> bool:
        """Recompute & apply every monitors's layout."""
        if isinstance(monitors, str):
            self.log.error("relayout doesn't take any argument")
            await self.notify_error("relayout doesn't take any argument")
            return False

        self._clear_mon_by_pat_cache()

        if monitors is None:
            monitors = cast(list[MonitorInfo], await self.hyprctl_json("monitors"))

        cleaned_config = self.resolve_names(monitors)
        if cleaned_config:
            self.log.debug("Using %s", cleaned_config)
        else:
            self.log.debug("No configuration item is applicable")
            return False
        graph = build_graph(cleaned_config)
        need_change = self._update_positions(monitors, graph, cleaned_config)
        every_monitor = {v["name"]: v for v in await self.hyprctl_json("monitors all")}
        if self.cast_bool(self.config.get("trim_offset"), True):
            trim_offset(monitors)

        for monitor in sorted(monitors, key=lambda x: x["x"] + x["y"]):
            await self.hyprctl(self._build_monitor_command(monitor, cleaned_config, every_monitor), "keyword")
        return need_change

    # Event handlers

    async def event_monitoradded(self, name: str) -> None:
        """Triggers when a monitor is plugged."""
        await asyncio.sleep(self.config.get("new_monitor_delay", 1.0))
        monitors = await self.hyprctl_json("monitors")
        await self._hotplug_command(monitors, name)

        if not await self._run_relayout(monitors):
            default_command = self.config.get("unknown")
            if default_command:
                await asyncio.create_subprocess_shell(default_command)

    # Utils

    async def _hotplug_command(self, monitors: list[MonitorInfo], name: str) -> None:
        """Run the hotplug command for the monitor."""
        monitors_by_descr = {m["description"]: m for m in monitors}
        monitors_by_name = {m["name"]: m for m in monitors}
        for descr, command in self.config.get("hotplug_commands", {}).items():
            mon = self._get_mon_by_pat(descr, monitors_by_descr, monitors_by_name)
            if mon and mon["name"] == name:
                await asyncio.create_subprocess_shell(command)
                break
        single_command = self.config.get("hotplug_command")
        if single_command:
            await asyncio.create_subprocess_shell(single_command)

    def _clear_mon_by_pat_cache(self) -> None:
        """Clear the cache."""
        self._mon_by_pat_cache: dict[str, MonitorInfo] = {}

    def _get_mon_by_pat(self, pat: str, description_db: dict[str, MonitorInfo], name_db: dict[str, MonitorInfo]) -> MonitorInfo | None:
        """Return a (plugged) monitor object given its pattern or none if not found."""
        cached = self._mon_by_pat_cache.get(pat)
        if cached is None:
            cached = name_db.get(pat)
            if cached is None:
                for full_descr in description_db:
                    if pat in full_descr:
                        cached = description_db[full_descr]
                        break
            if cached:
                self._mon_by_pat_cache[pat] = cached
        return cached

    _flipped_positions = {
        "topof": "bottomof",
        "bottomof": "topof",
        "leftof": "rightof",
        "rightof": "leftof",
        "topmiddleof": "bottommiddleof",
        "bottommiddleof": "topmiddleof",
        "leftmiddleof": "rightmiddleof",
        "rightmiddleof": "leftmiddleof",
        "topcenterof": "bottomcenterof",
        "bottomcenterof": "topcenterof",
        "leftcenterof": "rightcenterof",
        "rightcenterof": "leftcenterof",
        "topendof": "bottomendof",
        "bottomendof": "topendof",
        "leftendof": "rightendof",
        "rightendof": "leftendof",
    }

    def _update_positions(self, monitors: list[MonitorInfo], graph: dict[str, list[str]], config: dict[str, dict[str, list[str]]]) -> bool:
        """Apply configuration to monitors_by_name using graph."""
        monitors_by_name = {m["name"]: m for m in monitors}
        requires_update = False
        for _ in range(len(monitors_by_name) ** 2):
            changed = False
            for name in reversed(graph):
                mon1 = monitors_by_name[name]
                for name2 in graph[name]:
                    mon2 = monitors_by_name[name2]
                    for pos, _ in self.get_matching_config(name, name2, config):
                        try:
                            x, y = get_xy(self._flipped_positions[pos.lower()], mon2, mon1)
                        except TypeError:
                            self.log.exception("Invalid position %s", pos)
                            continue
                        if x != mon2["x"]:
                            changed = True
                            requires_update = True
                            mon2["x"] = x
                        if y != mon2["y"]:
                            changed = True
                            requires_update = True
                            mon2["y"] = y
            if not changed:
                break
        return requires_update

    def get_matching_config(self, name1: str, name2: str, config: dict[str, dict[str, list[str]]]) -> list[tuple[str, str]]:
        """Return rules matching name1 or name2 (relative to name1), looking up config.

        Returns a list of tuples (position, name) where name is the other monitor's name.
        """
        results = []
        ref_set = {name1, name2}
        for name_a, positions in config.items():
            for pos, names in positions.items():
                if pos in MONITOR_PROPS:
                    continue
                lpos = clean_pos(pos)
                for name_b in names:
                    if {name_a, name_b} == ref_set:
                        if name_a == name1:
                            results.append((lpos, name_b))
                        else:
                            results.append((self._flipped_positions[lpos], name_a))
        return results

    def resolve_names(self, monitors: list[MonitorInfo]) -> dict[str, Any]:
        """Change partial descriptions used in config for monitor names.

        Args:
            monitors: list of plugged monitors
        Returns:
            dict: cleaned config
        """
        monitors_by_descr = {m["description"]: m for m in monitors}
        cleaned_config: dict[str, dict[str, Any]] = {}
        plugged_monitors = {m["name"]: m for m in monitors}
        for descr1, placement in deepcopy(self.config.get("placement", {})).items():
            mon = self._get_mon_by_pat(descr1, monitors_by_descr, plugged_monitors)
            if not mon:
                continue
            name = mon["name"]
            if name not in plugged_monitors:
                continue
            cleaned_config[name] = {}
            for position, descr_list in placement.items():
                if position in MONITOR_PROPS:
                    cleaned_config[name][position] = descr_list
                else:
                    if not isinstance(descr_list, list | str):
                        errmsg = f'Unexpected monitor setting: {position}: "{descr_list}"'
                        raise ValueError(errmsg)
                    resolved = []
                    for props in [descr_list] if isinstance(descr_list, str) else descr_list:
                        info = self._get_mon_by_pat(props, monitors_by_descr, plugged_monitors)
                        if info:
                            resolved.append(info["name"])
                    if resolved:
                        cleaned_config[name][clean_pos(position)] = [r for r in resolved if r in plugged_monitors]
        return cleaned_config



"""Utilities."""

__all__ = ["CacheData"]

import asyncio
import time
from dataclasses import dataclass, field

from .types import JSONResponse


@dataclass
class CacheData:
    """Cache data structure."""

    retension_time: float
    expiration_date: float = 0
    payload: None | JSONResponse = field(default_factory=dict)
    _signal: asyncio.Event = field(default_factory=asyncio.Event)

    def set_pending(self, ref_time: float | None = None) -> None:
        """Set the current coroutine which will return the result.

        Will mark data as not ready yet, blocking awaiters of `wait_update`
        """
        self._signal.clear()
        self.expiration_date = (ref_time or time.time()) + self.retension_time
        self.payload = None

    def set_value(self, value: JSONResponse) -> None:
        """Set the cached value.

        Unblocks awaiters of `wait_update`
        """
        self.payload = value
        self._signal.set()

    async def wait_update(self) -> JSONResponse:
        """Wait for the cache data to be refreshed."""
        while True:
            if self.payload is not None:
                return self.payload
            await self._signal.wait()



"""Placement for absolute window positioning."""

__all__ = ["Placement"]

import enum
from typing import cast

from ...adapters.units import convert_monitor_dimension
from ...types import ClientInfo, MonitorInfo
from .helpers import get_size


class AnimationTarget(enum.Enum):
    """Animation type (selects between main window and satellite windows)."""

    MAIN = "main"
    EXTRA = "extra"
    ALL = "all"


class Placement:  # {{{
    """Animation store."""

    # main function
    @staticmethod
    def get(animation_type: str, monitor: MonitorInfo, client: ClientInfo, margin: int) -> tuple[int, int]:
        """Get destination coordinate for the provided animation type."""
        return cast(tuple[int, int], getattr(Placement, animation_type)(monitor, client, margin))

    # animation types
    @staticmethod
    def fromtop(monitor: MonitorInfo, client: ClientInfo, margin: int) -> tuple[int, int]:
        """Slide from/to top."""
        mon_x = monitor["x"]
        mon_y = monitor["y"]
        mon_width, mon_height = get_size(monitor)

        client_width = client["size"][0]
        margin_x = int((mon_width - client_width) / 2) + mon_x

        corrected_margin = convert_monitor_dimension(margin, mon_height, monitor)

        return (margin_x, mon_y + corrected_margin)

    @staticmethod
    def frombottom(monitor: MonitorInfo, client: ClientInfo, margin: int) -> tuple[int, int]:
        """Slide from/to bottom."""
        mon_x = monitor["x"]
        mon_y = monitor["y"]
        mon_width, mon_height = get_size(monitor)

        client_width = client["size"][0]
        client_height = client["size"][1]
        margin_x = int((mon_width - client_width) / 2) + mon_x

        corrected_margin = convert_monitor_dimension(margin, mon_height, monitor)

        return (margin_x, mon_y + mon_height - client_height - corrected_margin)

    @staticmethod
    def fromleft(monitor: MonitorInfo, client: ClientInfo, margin: int) -> tuple[int, int]:
        """Slide from/to left."""
        mon_x = monitor["x"]
        mon_y = monitor["y"]
        mon_width, mon_height = get_size(monitor)

        client_height = client["size"][1]
        margin_y = int((mon_height - client_height) / 2) + mon_y

        corrected_margin = convert_monitor_dimension(margin, mon_width, monitor)

        return (corrected_margin + mon_x, margin_y)

    @staticmethod
    def fromright(monitor: MonitorInfo, client: ClientInfo, margin: int) -> tuple[int, int]:
        """Slide from/to right."""
        mon_x = monitor["x"]
        mon_y = monitor["y"]
        mon_width, mon_height = get_size(monitor)

        client_width = client["size"][0]
        client_height = client["size"][1]
        margin_y = int((mon_height - client_height) / 2) + mon_y

        corrected_margin = convert_monitor_dimension(margin, mon_width, monitor)

        return (mon_width - client_width - corrected_margin + mon_x, margin_y)


# }}}



"""Force workspaces to follow the focus / mouse."""

from typing import cast

from .interface import Plugin


class Extension(Plugin):  # pylint: disable=missing-class-docstring
    """Force workspaces to follow the focus / mouse."""

    workspace_list: list[int] = []

    async def on_reload(self) -> None:
        """Rebuild workspaces list."""
        self.workspace_list = list(range(1, self.config.get("max_workspaces", 10) + 1))

    async def event_focusedmon(self, screenid_name: str) -> None:
        """Reacts to monitor changes."""
        monitor_id, workspace_name = screenid_name.split(",")
        # move every free workspace to the currently focused desktop
        busy_workspaces = {
            mon["activeWorkspace"]["name"] for mon in cast(list[dict], await self.hyprctl_json("monitors")) if mon["name"] != monitor_id
        }
        workspaces = [w["name"] for w in cast(list[dict], await self.hyprctl_json("workspaces")) if w["id"] > 0]

        batch: list[str] = []
        for n in workspaces:
            if n in busy_workspaces or n == workspace_name:
                continue
            batch.append(f"moveworkspacetomonitor name:{n} {monitor_id}")
        await self.hyprctl(batch)

    async def run_change_workspace(self, direction: str) -> None:
        """<+1/-1> Switch workspaces of current monitor, avoiding displayed workspaces."""
        increment = int(direction)
        # get focused screen info
        monitors = await self.hyprctl_json("monitors")
        assert isinstance(monitors, list)
        for monitor in monitors:
            if monitor["focused"]:
                break
        else:
            self.log.error("Can not find a focused monitor")
            return
        assert isinstance(monitor, dict)
        busy_workspaces = {m["activeWorkspace"]["id"] for m in monitors if m["id"] != monitor["id"]}
        cur_workspace = monitor["activeWorkspace"]["id"]
        available_workspaces = [i for i in self.workspace_list if i not in busy_workspaces]
        try:
            idx = available_workspaces.index(cur_workspace)
        except ValueError:
            next_workspace = available_workspaces[0 if increment > 0 else -1]
        else:
            next_workspace = available_workspaces[(idx + increment) % len(available_workspaces)]
        await self.hyprctl(
            [
                f"moveworkspacetomonitor {next_workspace},{monitor['name']}",
                f"workspace {next_workspace}",
            ]
        )



"""Moves unreachable client windows to the currently focused workspace."""

from typing import Any, cast

from ..types import ClientInfo, MonitorInfo
from .interface import Plugin


def contains(monitor: MonitorInfo, window: ClientInfo) -> bool:
    """Tell if a window is visible in a monitor."""
    if not (window["at"][0] > monitor["x"] and window["at"][0] < monitor["x"] + monitor["width"]):
        return False
    return bool(window["at"][1] > monitor["y"] and window["at"][1] < monitor["y"] + monitor["height"])


class Extension(Plugin):  # pylint: disable=missing-class-docstring
    """Moves unreachable client windows to the currently focused workspace."""

    async def run_attract_lost(self) -> None:
        """Brings lost floating windows to the current workspace."""
        monitors = cast(list, await self.hyprctl_json("monitors"))
        windows = cast(list, await self.get_clients())
        lost = [win for win in windows if win["floating"] and not any(contains(mon, win) for mon in monitors)]
        focused: dict[str, Any] = next(mon for mon in monitors if mon["focused"])
        interval = focused["width"] / (1 + len(lost))
        interval_y = focused["height"] / (1 + len(lost))
        batch = []
        workspace: int = focused["activeWorkspace"]["id"]
        margin = interval // 2
        margin_y = interval_y // 2
        for i, window in enumerate(lost):
            pos_x = int(margin + focused["x"] + i * interval)
            pos_y = {int(margin_y + focused["y"] + i * interval_y)}
            batch.append(f'movetoworkspacesilent {workspace},pid:{window["pid"]}')
            batch.append(f'movewindowpixel exact {pos_x} {pos_y},pid:{window["pid"]}')
        await self.hyprctl(batch)



"""expose Brings every client window to screen for selection."""

from ..common import CastBoolMixin, state
from ..types import ClientInfo
from .interface import Plugin


class Extension(CastBoolMixin, Plugin):  # pylint: disable=missing-class-docstring
    """Expose all clients on the active workspace."""

    exposed: list[ClientInfo] = []

    @property
    def exposed_clients(self) -> list[ClientInfo]:
        """Returns the list of clients currently using exposed mode."""
        if self.cast_bool(self.config.get("include_special"), False):
            return self.exposed
        return [c for c in self.exposed if c["workspace"]["id"] > 0]

    async def run_expose(self) -> None:
        """Expose every client on the active workspace.

        If expose is active restores everything and move to the focused window
        """
        if self.exposed:
            commands = [
                f"movetoworkspacesilent {client['workspace']['name']},address:{client['address']}" for client in self.exposed_clients
            ]
            commands.extend(
                (
                    "togglespecialworkspace exposed",
                    f"focuswindow address:{state.active_window}",
                )
            )
            await self.hyprctl(commands)
            self.exposed = []
        else:
            self.exposed = await self.get_clients(workspace_bl=state.active_workspace)
            commands = []
            for client in self.exposed_clients:
                commands.append(f"movetoworkspacesilent special:exposed,address:{client['address']}")
            commands.append("togglespecialworkspace exposed")
            await self.hyprctl(commands)



"""Shift workspaces across monitors."""

from typing import cast

from .interface import Plugin


class Extension(Plugin):  # pylint: disable=missing-class-docstring
    """Shift workspaces across monitors."""

    monitors: list[str] = []

    async def init(self) -> None:
        """Initialize the plugin."""
        self.monitors: list[str] = [mon["name"] for mon in cast(list[dict], await self.hyprctl_json("monitors"))]

    async def run_shift_monitors(self, arg: str) -> None:
        """<+1/-1> Swaps monitors' workspaces in the given direction."""
        direction: int = int(arg)
        mon_list = self.monitors[:-1] if direction > 0 else list(reversed(self.monitors[1:]))

        for i, mon in enumerate(mon_list):
            await self.hyprctl(f"swapactiveworkspaces {mon} {self.monitors[i + direction]}")

    async def event_monitoradded(self, monitor: str) -> None:
        """Keep track of monitors."""
        self.monitors.append(monitor)

    async def event_monitorremoved(self, monitor: str) -> None:
        """Keep track of monitors."""
        self.monitors.remove(monitor)



"""Common plugin interface."""

import contextlib
from collections.abc import Callable
from typing import Any, cast

from ..common import get_logger
from ..ipc import get_controls
from ..types import ClientInfo


class Plugin:
    """Base class for any pyprland plugin."""

    aborted = False

    hyprctl_json: Callable
    " `pyprland.ipc.hyprctl_json` using the pluggin's logger "

    hyprctl: Callable
    " `pyprland.ipc.hyprctl` using the pluggin's logger "

    notify: Callable
    " `pyprland.ipc.notify` using the pluggin's logger "

    notify_info: Callable
    " `pyprland.ipc.notify_info` using the pluggin's logger "

    notify_error: Callable
    " `pyprland.ipc.notify_error` using the pluggin's logger "

    config: dict[str, Any]
    " This plugin configuration section as a `dict` object "

    def __init__(self, name: str) -> None:
        """Create a new plugin `name` and the matching logger."""
        self.name = name
        """ the plugin name """
        self.log = get_logger(name)
        """ the logger to use for this plugin """
        ctrl = get_controls(self.log)
        (
            self.hyprctl,
            self.hyprctl_json,  # pylint: disable=invalid-name
            self.notify,
            self.notify_info,
            self.notify_error,
        ) = ctrl
        self.config = {}

    # Functions to override

    async def init(self) -> None:
        """Initialize the plugin.

        Note that the `config` attribute isn't ready yet when this is called.
        """

    async def on_reload(self) -> None:
        """Add the code which requires the `config` attribute here.

        This is called on *init* and *reload*
        """

    async def exit(self) -> None:
        """Empty exit function."""

    # Generic implementations

    async def load_config(self, config: dict[str, Any]) -> None:
        """Load the configuration section from the passed `config`."""
        self.config.clear()
        with contextlib.suppress(KeyError):
            self.config.update(config[self.name])

    async def get_clients(self, mapped: bool = True, workspace: None | str = None, workspace_bl: str | None = None) -> list[ClientInfo]:
        """Return the client list, optionally returns only mapped clients or from a given workspace."""
        return [
            client
            for client in cast(list[ClientInfo], await self.hyprctl_json("clients"))
            if (not mapped or client["mapped"])
            and (workspace is None or cast(str, client["workspace"]["name"]) == workspace)
            and (workspace_bl is None or cast(str, client["workspace"]["name"]) != workspace_bl)
        ]
